package MARC::NDL_Search::OPAC;

use 5.008008;
use strict;
use warnings;

require Exporter;

our @ISA = qw(Exporter);
our $VERSION = '0.01';

use LWP::UserAgent;
use HTTP::Status;
use URI::Escape;
use XML::Simple;
use Business::ISBN;

# package costractor
sub new {
    my $proto = shift;
    my $query = shift;
    my $class = ref($proto) || $proto;
    my $self = {};
    bless ($self, $class);
    set_query($self, $query);
    return $self;
}

# 検索式をセット
sub set_query {
    my $self = shift;
    my $query = shift;
    $self->{'url'}  = 'http://iss.ndl.go.jp/api/opensearch?dpid=iss-ndl-opac';
    $self->{'max'}       = $query->{'max'}       || 20;
    $self->{'page'}      = $query->{'page'}      || 1;
    $self->{'timeout'}   = $query->{'timeout'}   || 60;
    $self->{'any'}       = $query->{'any'}       || '';
    $self->{'title'}     = $query->{'title'}     || '';
    $self->{'creator'}   = $query->{'creator'}   || '';
    $self->{'publisher'} = $query->{'publisher'} || '';
    $self->{'ndc'}       = $query->{'ndc'}       || '';
    $self->{'isbn'}      = $query->{'isbn'}      || '';
    $self->{'from'}      = $query->{'from'}      || '';
    $self->{'until'}     = $query->{'until'}     || '';

    $self->{'any'}     =~ s/　/ /g;
    $self->{'title'}   =~ s/　/ /g;
    $self->{'creator'} =~ s/　/ /g;
    $self->{'isbn'}    =~ s/\-//g;

    # URI Encode
    $self->{'any'}       = uri_escape($self->{'any'});
    $self->{'title'}     = uri_escape($self->{'title'});
    $self->{'creator'}   = uri_escape($self->{'creator'});
    $self->{'publisher'} = uri_escape($self->{'publisher'});
}

# 検索ヒット件数を返す
sub hits {
    my $self = shift;
    return $self->{'hits'};
}

# 検索結果一覧（配列レファレンス）を返す
sub records {
    my $self = shift;
    return $self->{'records'};
}

# httpステータスコードほかのエラーを返す
sub error {
    my $self = shift;
    return $self->{'error'};
}

# 国会図書館サーチを実行
sub search {
    my $self = shift;
    my $key  = shift;
    my $biblio = {};
    my @biblios;

    my $url = $self->{'url'}
             . '&cnt='        . $self->{'max'}
             . '&idx='        . $self->{'page'}
             . '&any='        . $self->{'any'}
             . '&title='      . $self->{'title'}
             . '&creator='    . $self->{'creator'}
             . '&publisher='  . $self->{'publisher'}
             . '&ndc='        . $self->{'ndc'}
             . '&isbn='       . $self->{'isbn'}
             . '&from='       . $self->{'from'}
             . '&until='      . $self->{'until'}
             ;
    my ($input, $error) = _access_web($url);
    $self->{'error'} = $error;

    if ($input) {
        my $xml_ref = XMLin(\$input);
        my $hits  = $xml_ref->{'channel'}{'openSearch:totalResults'};
        $self->{'hits'} = $hits;
        return $self if $hits == 0;

        my $items = _repeat_field($xml_ref->{'channel'}{'item'});
        foreach my $item (@{$items}) {
            $biblio = _get_bibliography($item, $hits);
            push @biblios,  $biblio;
        }
    }
    $self->{'records'} = \@biblios;
    return $self;
}

# 国会図書館ISBNサーチを実行
sub isbn_search {
    my $self = shift;
    my $isbn = shift;
    $isbn =~ s/\-//g;
    my $biblio = {};

    my $url = $self->{'url'} . '&isbn=' . $isbn;
    my ($input, $error) = _access_web($url, $self->{'timeout'});
    $self->{'error'} = $error if $error;

    if ($input) {
        my $xml_ref = XMLin(\$input);
        my $hits  = $xml_ref->{'channel'}{'openSearch:totalResults'};
        $self->{'hits'} = $hits;
        if ($hits == 0) {
            return $self;
        }
        my $item;
        if ($hits > 1) {
            $item = shift @{$xml_ref->{'channel'}{'item'}};
        }
        else {
           $item = $xml_ref->{'channel'}{'item'};
        }
        $biblio = _get_bibliography($item, $hits);
    }
    else {
        $self->{'hits'} = 0;
    }
    $self->{'records'} = $biblio;
    return $self;
}

# 検索結果から書誌を取り出す
sub _get_bibliography {
   my $item = shift;
   my $hits = shift;

   my %biblio = (
       'title'        => "",    # タイトル
       'title_yomi'   => "",    # タイトル読み
       'series'       => "",    # シリーズ
       'series_yomi'  => "",    # シリーズ読み
       'author'       => [],    # 著者(繰り返し項目)
       'author_yomi'  => [],    # 著者読み（繰り返し項目)
       'publisher'    => [],    # 出版者(繰り返し項目)
       'link'         => "",    # リンク
       'category'     => "",    # 分野
       'description'  => "",    # 記述
       'note'         => [],    # 注記
       'pubDate'      => "",    # 出版日
       'guid'         => "",    # guid
       'ISBN'         => "",    # ISBN
       'NDLC'         => "",    # NDLC
       'NDC9'         => "",    # NDC9
       'subject'      => "",    # 件名
    );

    my $identifers = _repeat_field($item->{'dc:identifier'});
    foreach my $identifer (@{$identifers}){
         if (defined $identifer->{'xsi:type'} &&
             $identifer->{'xsi:type'} eq 'dcndl:ISBN') {
                 $biblio{"ISBN"} = $identifer->{'content'}
         }
    }

    $biblio{'title'}       = $item->{'title'};
    $biblio{'title_yomi'}  = $item->{'dcndl:titleTranscription'};
    $biblio{'volume'}      = $item->{'dcndl:volume'};
    $biblio{'author'}      = _repeat_field($item->{'dc:creator'});
    $biblio{'author_yomi'} = _repeat_field($item->{'dcndl:creatorTranscription'});
    $biblio{'link'}        = $item->{'link'};
    $biblio{'description'} = $item->{'description'};
    $biblio{'category'}    = $item->{'category'};
    $biblio{'guid'}        = $item->{'guid'}{'content'};
    $biblio{'pubDate'}     = $item->{'pubDate'};
    $biblio{'publisher'}   = _repeat_field($item->{'dc:publisher'});
    $biblio{'note'}        = _repeat_field($item->{'dcterms:description'});

    my $subject_ref = [];
    if (ref $item->{'dc:subject'} ne 'ARRAY') {
        push @{$subject_ref}, $item->{'dc:subject'};
    }
    else {
        $subject_ref = $item->{'dc:subject'};
    }

    foreach my $dc_subject (@{$subject_ref}) {
        if (ref $dc_subject eq 'HASH') {
             $biblio{'NDLC'} = $dc_subject->{'content'}
                 if $dc_subject->{'xsi:type'} eq 'dcndl:NDLC';
             $biblio{'NDC9'} = $dc_subject->{'content'}
                 if $dc_subject->{'xsi:type'} eq 'dcndl:NDC9';
        } 
        else {
             $biblio{'subject'} = $dc_subject;
        }
    }

    return \%biblio;
}

# Scalar値と配列リファレンスを配列リファレンスに統一
sub _repeat_field {
    my $orgdata    = shift;
    if (ref $orgdata eq "ARRAY") {
        return $orgdata;
    }
    eles {
        return [$orgdata];
    }
}

# Webサーバへのアクセス
sub _access_web {
    my $url = shift;
    my $timeout = shift;
    my $usa = new LWP::UserAgent;
    my $request = new HTTP::Request('GET', $url);
    $usa->timeout([$timeout]);

    # BODY情報を得る
    my $response = $usa->request($request);
    if ($response->code != RC_OK) {
        return (undef, $response->code);
    }
    else {
        return ($response->content, undef);
    }
}

1;

__END__

=head1 NAME

MARC::NDL_Search::OPAC - 国立国会図書館サーチＡＰＩにおけるOPAC検索のPerlモジュール

=head1 概要

=head2 オブジェクトの作成

  use MARC::NDL_Search::OPAC;
  my $NDL_Search = new MARC::NDL_Search::OPAC;

=head2 検索の実行

=head2 キーワード検索

  $NDL_Search->set_query({'any' => '吾輩は猫である'}); # キーワード検索
  $NDL_Search->search;
  my $records = $NDL_sarch->reords;
  foreach $biblio (@{$records) {
      print "タイトル: ", $biblio->{"title"}, "\n";
      print "著者: ",     join q{, }, @{$biblio->{"author"}}, "\n";
      print "出版者: ",   join q{, }, @{$biblio->{"publisher"}}, "\n";
  }

=head2 ISBN検索

  $NDL_Search->isbn_search('4101010013'); 
  print "タイトル: ", $biblio->{"title"}, "\n";
  print "著者: ",     join q{, }, @{$biblio->{"author"}}, "\n";
  print "出版者: ",   join q{, }, @{$biblio->{"publisher"}}, "\n";

=head2 エラーとヒット件数の取得

=head3  エラーの補足

  my $error = $NDL_Search->error;
  exit if ($error);

=head3 ヒット件数の取得

  my $hits  = $NDL_Search->hists;


=head1 説明

国会図書館サーチ APIのNDL-OPACの利用に特化したPerlモジュールです。
国会図書館の蔵書をこのモジュール経由で検索することができます。


=head2 キーワード検索

=head3 キーワード検索のパラメータ指定

  検索キーワードや表示パラメータは、ハッシュの形式で渡します。
  キーワードはutf8文字列で与える必要があります。

  my %query = (
      'any'       => $any,        # キーワード検索
      'title'     => $title,      # タイトル検索
      'creator'   => $creator,    # 著者検索
      'publisher' => $publisher,  # 出版社検索
      'from'      => $from,       # 刊行日FORM (YYYY-MM-DD)
      'until'     => $until,      # 刊行日UNTIL(YYYY-MM-DD)
      'ndc'       => $ndc,        # NDC
      'isbn'      => $isbn,       # ISBN検索
      'max'       => $max,        # ページごと表示件数
      'page'      => $page,       # ページ番号
      'timeout'   => $timeout,    # タイムアウト
  );
  $NDL_Search->set_query(\%query);
  $NDL_Search->search;

=head3 キーワード検索の結果表示

  キーワード検索の結果は書誌レコードの配列レファレンスになります。
  各レコード中の項目は、書誌レコード（ハッシュレファレンス）から
取り出すことができます。
　いくつかの項目は、配列（繰り返し項目）になっています。
　呼び出せるデータ項目と、どの項目が繰り返し項目かは以下のコードで
確認ください。

  my $records = $NDL_Search->records;
  exit if $NDL_Search->error;
  if (!$NDL_Search->hits) {
      foreach my $biblio (@{$records}) {
        print "タイトル: ",       $biblio->{"title"}, "\n";
        print "タイトル(読み): ", $biblio->{"title_yomi"}, "\n";
        print "シリーズ: ",       $biblio->{"series"}, "\n";
        print "シリーズ(読み): ", $biblio->{"series_yomi"}, "\n";
        print "巻: ",             $biblio->{"volume"}, "\n";
        print "著者: ",           join q{, }, @{$biblio->{"author"}}, "\n";
        print "著者(読み): ",     join q{, }, @{$biblio->{"author_yomi"}}, "\n";
        print "出版者: ",         join q{, }, @{$biblio->{"publisher"}}, "\n";
        print "カテゴリー: ",     $biblio->{"category"}, "\n";
        print "注記: ",           join q{, }, @{$biblio->{"note"}}, "\n";
        print "guid: ",           $biblio->{"guid"}, "\n";
        print "出版日: ",         $biblio->{"pubDate"}, "\n";
        print "description: ",    $biblio->{"description"}, "\n";
        print "ISBN: ",           $biblio->{"ISBN"}, "\n";
        print "NDLC: ",           $biblio->{"NDLC"}, "\n";
        print "NDC9: ",           $biblio->{"NDC9"}, "\n";
        print "subject: ",        $biblio->{"subject"}, "\n";
        print "\n\n";
      }
  }

=head2 ISBN検索

  ISBNをキーに１件のデータのみ取り出します。日本刊行書籍限定です。
なお、このisbn_searchメソッドでは、ISBNの13桁・10桁変換を自動的に行います。
  キーワード検索と異なり、検索結果は書誌レコードの配列ではなく、１件の
書誌レコードとして戻ります。
  書誌レコードのデータ形式は、キーワード検索と同じです。

  $NDL_Search->isbn_search($isbn);
  my $biblio = $NDL_Serach->records;
  exit if $NDL_Search->error;
  if ($NDL_Search->hits > 0) {
        print "タイトル: ",       $biblio->{"title"}, "\n";
        print "タイトル(読み): ", $biblio->{"title_yomi"}, "\n";
        print "シリーズ: ",       $biblio->{"series"}, "\n";
        print "シリーズ(読み): ", $biblio->{"series_yomi"}, "\n";
        print "巻: ",             $biblio->{"volume"}, "\n";
        print "著者: ",           join q{, }, @{$biblio->{"author"}}, "\n";
        print "著者(読み): ",     join q{, }, @{$biblio->{"author_yomi"}}, "\n";
        print "出版者: ",         join q{, }, @{$biblio->{"publisher"}}, "\n";
        print "カテゴリー: ",     $biblio->{"category"}, "\n";
        print "注記: ",           join q{, }, @{$biblio->{"note"}}, "\n";
        print "guid: ",           $biblio->{"guid"}, "\n";
        print "出版日: ",         $biblio->{"pubDate"}, "\n";
        print "description: ",    $biblio->{"description"}, "\n";
        print "ISBN: ",           $biblio->{"ISBN"}, "\n";
        print "NDLC: ",           $biblio->{"NDLC"}, "\n";
        print "NDC9: ",           $biblio->{"NDC9"}, "\n";
        print "subject: ",        $biblio->{"subject"}, "\n";
  }

=head2 エクスポート

  このモジュールは何もエクスポートしません。

=head2 依存モジュール

  事前にPerlモジュールBusiness::ISBNをインストールしてください。
  それ以外は、Perlの標準的なモジュールを使用しています。

=head2 免責事項

  このモジュールのご利用に際して発生した不利益にいては、
一切責任を負いかねますので、ご了解の上で利用ください。

=head1 AUTHOR

Akira MaedaE<lt>maeda@lib.u-tokyo.ac.jpE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2011 by Akira Maeda

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.8 or,
at your option, any later version of Perl 5 you may have available.

=cut

